/*
    Copyright (C) 2010-2011 wpitchoune@gmail.com

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
*/

/* to use the GNU version of basename */
#ifndef _GNU_SOURCE
#define _GNU_SOURCE 1
#endif

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <getopt.h>

/* Directory separator is \ when cross-compiling for MS Windows
   systems */
#if defined(__MINGW32__)
#define DIRSEP '\\'
#else
#define DIRSEP '/'
#endif

#ifndef MP3TOOL_VERSION
#define MP3TOOL_VERSION "0.0.0"
#endif

#define PROGRAM_NAME "mp3tool"

#define PROGRAM_USAGE "mp3tool [OPTION] [mp3_path]"

#define PROGRAM_DESCRIPTION "Command line program for moving a MP3 file or\
 displaying its information."

#define PROGRAM_HELP "  --help    display this help and exit\n\
  --version   output version information and exit\n\
  -m PATTERN, --move=PATTERN   Moves MP3 file to 'PATTERN'\n\
  -i PATTERN, --info=PATTERN   Displays MP3 information formatted according to\
 'PATTERN'\n\n\
PATTERN variables:\n\
  %f\tOriginal filename\n\
  %c\tMP3 comment\n\
  %a\tMP3 artist\n\
  %l\tMP3 album\n\
  %t\tMP3 title\n\
  %%\tPercent\n\
  %n\tNewline\n\n\
Examples:\n\n\
  moves 'file.mp3' to 'artist_name/album_name/title_name.mp3':\n\n\
  mp3tool -m \"%a/%l/%t.mp3\" file.mp3\n\
"


#define PROGRAM_CONTACT "Report bugs to <wpitchoune@gmail.com>\n\
Home page: <http://wpitchoune.net/mp3tool/>"

#define PROGRAM_COPYRIGHT "Copyright (C) 2010-2011 wpitchoune@gmail.com.\n\
License GNU GPL version 2 \
<http://www.gnu.org/licenses/old-licenses/gpl-2.0.html>.\n\
This is free software: you are free to change and redistribute it.\n\
There is NO WARRANTY, to the extent permitted by law."



struct mp3_info {
	char title[31];
	char artist[31];
	char album[31];
	char year[5];
	char comment[31];
	char genre;
	char *filename;
};

/*
  Normalize an information field.

  If the string only contains whitespaces the string is replaced by
  'Unknown' string.
 */
void normalize(char *str)
{
	int unknown = 1;

	char *c = str;
	while (*c) {
		if (*c != ' ') {
			unknown = 0;
			break;
		}
		c++;
	}

	if (unknown)
		strcpy(str, "Unknown");
}

int read_mp3_info(const char *fpath, struct mp3_info *info)
{
	FILE *file;
	char tag[3];

	info->filename = strdup(basename(fpath));

	file = fopen(fpath, "r");

	if (!file)
		return 0;

	if (fseek(file, -128, SEEK_END) != 0)
		return 0;

	if (fread(&tag, 1, 3, file) != 3
	    || tag[0] != 'T' || tag[1] != 'A' || tag[2] != 'G')
		return 0;

	if (fread(&info->title, 1, 30, file) != 30)
		return 0;
	info->title[30] = '\0';
	normalize(info->title);

	if (fread(&info->artist, 1, 30, file) != 30)
		return 0;
	info->artist[30] = '\0';
	normalize(info->artist);

	if (fread(&info->album, 1, 30, file) != 30)
		return 0;
	info->album[30] = '\0';
	normalize(info->album);

	if (fread(&info->year, 1, 4, file) != 4)
		return 0;
	info->year[4] = '\0';

	if (fread(&info->comment, 1, 30, file) != 30)
		return 0;
	info->comment[30] = '\0';

	if (fread(&info->genre, 1, 1, file) != 1)
		return 0;

	fclose(file);

	return 1;
}

#define OPT_INFO "info"
#define OPT_MV "move"
#define OPT_HELP "help"
#define OPT_VERSION "version"

static struct option long_options[] = {
	{OPT_VERSION, no_argument, 0, 'v'},
	{OPT_HELP, no_argument, 0, 'h'},
	{OPT_INFO, required_argument, 0, 'i'},
	{OPT_MV, required_argument, 0, 'm'},
	{0, 0, 0, 0}
};


char *apply_pattern(const char *pattern, struct mp3_info *info)
{
	char *result = malloc(1);
	char *p;

	*result = '\0';
	p = (char *)pattern;

	while (*p) {
		char *old = result;
		int n;

		if (*p == '%' && *(p + 1)) {
			char *str = NULL;
			switch (*(p + 1)) {
			case 'a':
				str = info->artist;
				break;
			case 'l':
				str = info->album;
				break;
			case 'f':
				str = info->filename;
				break;
			case 't':
				str = info->title;
				break;
			case 'c':
				str = info->comment;
				break;
			case 'y':
				str = info->year;
				break;
			case 'n':
				str = "\n";
				break;
			case '%':
				str = "%";
				break;
			}

			if (str) {
				n = strlen(old) + strlen(str) + 1;
				result = malloc(n);
				strcpy(result, old);
				strcat(result, str);
				p++;
			} else {
				old = NULL;
			}
		} else {
			n = strlen(old) + 1 + 1;

			result = malloc(n);

			strcpy(result, old);
			result[n - 2] = *p;
			result[n - 1] = '\0';
		}

		free(old);

		p++;
	}

	return result;
}

void mkdirs(const char *dirs)
{
	char *c = (char *)dirs;
	char *dir = malloc(strlen(dirs) + 1);

	int i = 0;
	while (*c) {
		if ((*c == DIRSEP || *c == '\0') && c != dirs) {
			strncpy(dir, dirs, i);
			dir[i] = '\0';
			mkdir(dir, 0777);
		}

		c++;
		i++;
	}
	free(dir);

	mkdir(dirs, 0777);
}

char *wp_dirname(const char *path)
{
	int n = strlen(path);

	while (n >= 0) {
		if (path[n] == DIRSEP) {
			char *res = strdup(path);
			res[n] = '\0';
			return res;
		}

		n--;
	}

	return NULL;
}

int mp3_move(const char *fpath, const char *newpath_pattern)
{
	struct mp3_info info;
	char *npath, *dir;
	int ret = read_mp3_info(fpath, &info);

	if (!ret)
		return 0;

	npath = apply_pattern(newpath_pattern, &info);

	dir = wp_dirname(npath);

	if (dir)
		mkdirs(dir);

	printf("move: %s to %s\n", fpath, npath);
	ret = rename(fpath, npath);

	free(npath);
	free(dir);
	free(info.filename);

	return ret == 0;
}

int main(int argc, char **argv)
{
	char *path;
	char *mv_pattern = NULL;
	char *info_pattern = NULL;

	while (1) {
		int c, option_index = 0;

		c = getopt_long(argc, argv,
				"vhi:m:", long_options,
				&option_index);

		if (c == -1)
			break;

		switch (c) {

		case 0:
			break;

		case 'i':
			if (optarg)
				info_pattern = strdup(optarg);
			break;

		case 'm':
			if (optarg)
				mv_pattern = strdup(optarg);
			break;

		case 'h':
			printf("Usage: %s\n%s\n\nOptions:\n%s\n\n%s\n",
			       PROGRAM_USAGE,
			       PROGRAM_DESCRIPTION,
			       PROGRAM_HELP,
			       PROGRAM_CONTACT);
			exit(EXIT_SUCCESS);

		case 'v':
			printf("mp3tool %s\n", MP3TOOL_VERSION);
			printf("%s\n", PROGRAM_COPYRIGHT);

			exit(EXIT_SUCCESS);

		case '?':
			break;

		default:
			abort();
		}
	}

	if (optind + 1 != argc) {
		fprintf(stderr, "Invalid usage.\n");
		exit(EXIT_FAILURE);
	}

	path = argv[optind];

	if (info_pattern) {
		char *res;
		struct mp3_info info;

		if (!read_mp3_info(path, &info)) {
			fprintf(stderr,
				"ERROR: Failure to read MP3 information %s\n",
				path);
			exit(EXIT_FAILURE);
		}
		res = apply_pattern(info_pattern, &info);
		printf("%s", res);
		free(res);
		free(info.filename);
	}

	if (mv_pattern) {
		if (!mp3_move(path, mv_pattern)) {
			fprintf(stderr, "ERROR: Cannot move %s\n", path);
			exit(EXIT_FAILURE);
		}
	}

	free(info_pattern);
	free(mv_pattern);

	exit(EXIT_SUCCESS);
}
